--------------------------------------------------------------------------------
-- From Deriving to IG
--------------------------------------------------------------------------------

{-# LANGUAGE DataKinds              #-}
{-# LANGUAGE PolyKinds              #-}
{-# LANGUAGE TypeFamilies           #-}
{-# LANGUAGE TypeOperators          #-}
{-# LANGUAGE GADTs                  #-}
{-# LANGUAGE FlexibleInstances      #-}
{-# LANGUAGE FlexibleContexts       #-}
{-# LANGUAGE UndecidableInstances   #-}

module Conversion.D2IG where


import qualified Library.IG       as IG
import qualified Library.Deriving as D

--------------------------------------------------------------------------------
-- Conversion
--------------------------------------------------------------------------------

-- Type conversion
type family D2IG (u :: D.Un *) :: IG.Un *

type instance D2IG D.VV           = IG.VV
type instance D2IG D.UU           = IG.UU
type instance D2IG (D.KK i a)     = IG.KK a
type instance D2IG (D.MM i a)     = D2IG a
type instance D2IG ((D.:+:)  a b) = (IG.:+:)  (D2IG a) (D2IG b)
type instance D2IG ((D.:**:) a b) = (IG.:**:) (D2IG a) (D2IG b)

-- The remaining D representation types (PAR, REC, and :.:) are not considered
-- because we will only convert Generic representations, not Generic1.


-- Value conversion
class Convert (u :: D.Un *) where
  d2ig :: D.In u p -> IG.In (D2IG u)
  ig2d :: IG.In (D2IG u) -> D.In u p

instance Convert D.VV where
  d2ig = error "no values of type VV"
  ig2d = error "no values of type VV"

instance Convert D.UU where
  d2ig _ = IG.U
  ig2d _ = D.U1

instance (Convert a) => Convert (D.MM mc a) where
  d2ig (D.M1 a) = d2ig a
  ig2d a = D.M1 (ig2d a)

instance Convert (D.KK i a) where
  d2ig (D.K1 a) = IG.K a
  ig2d (IG.K a) = D.K1 a

instance (Convert a, Convert b) => Convert ((D.:+:) a b) where
  d2ig (D.L1 a) = IG.L (d2ig a)
  d2ig (D.R1 a) = IG.R (d2ig a)
  ig2d (IG.L a) = D.L1 (ig2d a)
  ig2d (IG.R a) = D.R1 (ig2d a)

instance (Convert a, Convert b) => Convert ((D.:**:) a b) where
  d2ig ((D.:*:) a b) = (IG.:*:) (d2ig a) (d2ig b)
  ig2d ((IG.:*:) a b) = (D.:*:) (ig2d a) (ig2d b)

--------------------------------------------------------------------------------
-- Embedding conversion
--------------------------------------------------------------------------------

-- All D types are representable in IG
instance (D.Generic a, Convert (D.Rep a)) => IG.Representable a where
  type Rep a = D2IG (D.Rep a)

  from x = d2ig (D.from x)
  to   x = D.to (ig2d x)

-- Note that this is the most general and only instance that can be given for
-- this class, but that's ok, because that's exactly what we want: never to have
-- to write instances for IG.Representable, simply use the stuff from D.Generic.
